"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DataSyncController = exports.assetFingerprint = void 0;
const axios_1 = __importDefault(require("axios"));
const util_1 = require("@cardano-graphql/util");
const cip14_js_1 = __importDefault(require("@emurgo/cip14-js"));
const errors_1 = require("./errors");
const object_hash_1 = __importDefault(require("object-hash"));
const ts_log_1 = require("ts-log");
exports.assetFingerprint = (asset) => new cip14_js_1.default(Buffer.from(asset.policyId, 'hex'), asset.assetName !== '' ? Buffer.from(asset.assetName, 'hex') : undefined)
    .fingerprint();
class DataSyncController {
    constructor(hasuraClient, db, pollingInterval, logger = ts_log_1.dummyLogger, metadataServerUri) {
        this.hasuraClient = hasuraClient;
        this.db = db;
        this.logger = logger;
        this.metadataServerUri = metadataServerUri;
        if (this.metadataServerUri) {
            this.axiosClient = axios_1.default.create({
                baseURL: metadataServerUri
            });
            this.metadataSynchronizer = {
                initial: new util_1.DataFetcher('MetadataSynchronizer', () => __awaiter(this, void 0, void 0, function* () {
                    const batchSize = 500;
                    const assetsWithoutMetadataCount = yield this.hasuraClient.assetsWithoutMetadataCount({ _lte: 2 });
                    this.logger.debug('Newly discovered assets missing metadata', { module: 'DataSyncController', value: assetsWithoutMetadataCount });
                    const batchQty = Math.ceil(assetsWithoutMetadataCount / batchSize);
                    let totalCount = 0;
                    for (const i of Array(batchQty).keys()) {
                        const assetsInBatch = yield this.hasuraClient.getAssetsWithoutMetadata({ _lte: 3 }, { limit: batchSize, offset: batchSize * i });
                        if (assetsInBatch.length > 0) {
                            yield this.fetchAndApplyMetadata(assetsInBatch);
                        }
                        totalCount = totalCount + assetsInBatch.length;
                    }
                    return totalCount;
                }), pollingInterval.initial, this.logger),
                ongoing: new util_1.DataFetcher('MetadataSynchronizerRefresh', () => __awaiter(this, void 0, void 0, function* () {
                    const batchSize = 500;
                    const assetsEligibleForMetadataRefreshCount = yield this.hasuraClient.assetsEligibleForMetadataRefreshCount({ _gt: 2 });
                    this.logger.debug('Assets eligible for metadata refresh', { module: 'DataSyncController', value: assetsEligibleForMetadataRefreshCount });
                    const batchQty = Math.ceil(assetsEligibleForMetadataRefreshCount / batchSize);
                    let totalCount = 0;
                    for (const i of Array(batchQty).keys()) {
                        const assetsInBatch = yield this.hasuraClient.getAssetsIncMetadata({ _gt: 2 }, { limit: batchSize, offset: batchSize * i });
                        if (assetsInBatch.length > 0) {
                            yield this.fetchAndApplyMetadata(assetsInBatch);
                        }
                        totalCount = totalCount + assetsInBatch.length;
                    }
                    return totalCount;
                }), pollingInterval.ongoing, this.logger)
            };
        }
        this.assetSynchronizer = new util_1.DataFetcher('AssetTableSynchronizer', () => __awaiter(this, void 0, void 0, function* () {
            const batchSize = 500;
            const distinctAssetsInTokensCount = yield this.hasuraClient.distinctAssetsInTokensCount();
            this.logger.debug('Distinct assets in tokens count', { module: 'DataSyncController', value: distinctAssetsInTokensCount });
            const batchQty = Math.ceil(distinctAssetsInTokensCount / batchSize);
            let totalCount = 0;
            for (const i of Array(batchQty).keys()) {
                const assetsInBatch = yield this.hasuraClient.getDistinctAssetsInTokens({ limit: batchSize, offset: batchSize * i });
                const assetsAlreadyInDb = yield this.hasuraClient.getAssetsById(assetsInBatch.map(asset => asset.assetId));
                this.logger.debug('Assets from tokens', {
                    module: 'DataSyncController',
                    value: { batch: i, qty: assetsInBatch.length, existing: assetsAlreadyInDb.length }
                });
                const newAssets = assetsInBatch
                    .filter(asset => assetsAlreadyInDb.find(existingAsset => existingAsset.assetId === asset.assetId) === undefined)
                    .map(asset => ({
                    assetId: asset.assetId,
                    assetName: asset.assetName,
                    fingerprint: exports.assetFingerprint(asset),
                    policyId: asset.policyId,
                    metadataFetchAttempts: 0
                }));
                this.logger.debug('asset IDs diff', { module: 'DataSyncController', value: newAssets.length });
                if (newAssets.length > 0) {
                    totalCount = totalCount + newAssets.length;
                    yield this.hasuraClient.insertAssets(newAssets);
                    this.logger.debug('synchronised assets table from tokens', { module: 'DataSyncController', value: { batch: i, qty: newAssets.length } });
                }
            }
            return totalCount;
        }), 120 * 1000, this.logger);
    }
    fetchAndApplyMetadata(assets) {
        return __awaiter(this, void 0, void 0, function* () {
            const newMetadata = yield this.getAssetMetadata(assets);
            const assetsWithMetadata = newMetadata
                .map(metadata => ({
                metadata,
                metadataHash: object_hash_1.default(metadata),
                asset: assets.find(asset => asset.assetId === metadata.subject)
            }))
                .filter(({ asset, metadataHash }) => metadataHash !== asset.metadataHash)
                .map(obj => {
                var _a, _b, _c, _d, _e;
                return (Object.assign(Object.assign(Object.assign({}, obj.asset), {
                    description: (_a = obj.metadata.description) === null || _a === void 0 ? void 0 : _a.value,
                    logo: (_b = obj.metadata.logo) === null || _b === void 0 ? void 0 : _b.value,
                    name: (_c = obj.metadata.name) === null || _c === void 0 ? void 0 : _c.value,
                    ticker: (_d = obj.metadata.ticker) === null || _d === void 0 ? void 0 : _d.value,
                    url: (_e = obj.metadata.url) === null || _e === void 0 ? void 0 : _e.value
                }), { metadataHash: object_hash_1.default(obj.metadata) }));
            });
            this.logger.debug('Metadata with updates to apply', { module: 'DataSyncController', value: assetsWithMetadata.length });
            if (assetsWithMetadata.length > 0) {
                yield this.hasuraClient.addMetadata(assetsWithMetadata);
            }
            yield this.hasuraClient.incrementMetadataFetchAttempts(assets.map(asset => asset.assetId));
        });
    }
    getAssetMetadata(assets) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const response = yield this.axiosClient.post('metadata/query', {
                    subjects: assets.map(asset => asset.assetId),
                    properties: [
                        'description',
                        'logo',
                        'name',
                        'ticker',
                        'url'
                    ]
                });
                return response.data.subjects;
            }
            catch (error) {
                if (error.code === 'ENOTFOUND') {
                    this.logger.error(error.message);
                }
                else {
                    throw error;
                }
            }
        });
    }
    ensureAssetFingerprints() {
        return __awaiter(this, void 0, void 0, function* () {
            const runBatch = (assets) => __awaiter(this, void 0, void 0, function* () {
                yield this.hasuraClient.addAssetFingerprints(assets.map((asset) => ({ assetId: asset.assetId, fingerprint: exports.assetFingerprint(asset) })));
            });
            while (yield this.hasuraClient.hasAssetsWithoutFingerprint()) {
                yield runBatch(yield this.hasuraClient.getAssetsWithoutFingerprint(2500));
            }
        });
    }
    ensureMetadataServerIsAvailable() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.axiosClient.get('/metadata/healthcheck');
            }
            catch (error) {
                if (error.code === 'ENOTFOUND') {
                    throw new errors_1.HostDoesNotExist('metadata server');
                }
                else if (error.response.status !== 404) {
                    throw error;
                }
            }
        });
    }
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.info('Initializing', { module: 'DataSyncController' });
            yield this.ensureAssetFingerprints();
            yield this.assetSynchronizer.initialize();
            if (this.metadataServerUri) {
                yield this.ensureMetadataServerIsAvailable();
                yield this.metadataSynchronizer.initial.initialize();
                yield this.metadataSynchronizer.ongoing.initialize();
            }
            this.logger.info('Initialized', { module: 'DataSyncController' });
        });
    }
    shutdown() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.info('Shutting down', { module: 'DataSyncController' });
            if (this.metadataServerUri) {
                yield this.metadataSynchronizer.initial.shutdown();
                yield this.metadataSynchronizer.ongoing.shutdown();
            }
            yield this.assetSynchronizer.shutdown();
            this.logger.info('Shutdown complete', { module: 'DataSyncController' });
        });
    }
}
exports.DataSyncController = DataSyncController;
//# sourceMappingURL=DataSyncController.js.map